/*********************************************************************/
/*      Licensed Materials - Property of IBM                         */
/*      "Restricted Materials of IBM"                                */
/*      5748-T15                                                     */
/*      (C) Copyright IBM Corp. 2006                                 */
/*      All Rights Reserved                                          */
/*      US Government Users Restricted Rights - Use, duplication     */
/*      or dislosure restricted by GSA ADP Schedule Contract         */
/*      with IBM Corp.                                               */
/*********************************************************************/
/* This file is provided on an "as is" basis.  There are no          */
/* warranties, express or implied, including the implied warranties  */
/* of merchantability and fitness for a particular purpose.          */
/* IBM has no obligation to provide service, defect correction, or   */
/* any maintenance for the file.  IBM has no obligation to supply    */
/* any updates or enhancements for the file to you even if such are  */
/* or later become available.                                        */
/*********************************************************************/
/* Change History                                                    */
/*   f11565     2/2/05     Use subsystem unique token for gsysc for  */
/*                         the SOLD driver control block.            */
/*   f9127      2/24/05    Capable of stopping SOLD in continuous    */
/*                         mode on a specific istream.               */
/*                                                                   */
/*                                                                   */
/*                                                                   */
/*********************************************************************/
/*                                                                   */
/* FUNCTION NAME.....QSCA                                            */
/* PURPOSE...........TPF Shared Object Linkage Driver                */
/* FUNCTION TYPE.....TPF                                             */
/* DOCUMENT NAME.....NA                                              */
/*                                                                   */
/* INPUT.......ZTEST i SOLD <START|CASE-xx|STOP>                     */
/*             OR ZTEST SOLD HELP or ? to display the correct        */
/*                syntax of the command.                             */
/*                                                                   */
/* Options:                                                          */
/* i : I-stream number (optional parameter)                          */
/* START: start all test cases in non-stop mode                      */
/* xx: is the test case number going to run                          */
/*                                                                   */
/* STOP: stop the driver                                             */
/* OUTPUT......Some programs are self checking the passing parameters*/
/*                                                                   */
/* DESCRIPTION:                                                      */
/*                                                                   */
/* QSCA is the "main" segment for the shared object linkage driver,  */
/* SOLD.  It determines which test case to run based on the number   */
/* entered by the user.                                              */
/*                                                                   */
/* Test Cases:                                                       */
/*                                                                   */
/* C language test cases: (test case slots 1-25)                     */
/*   1: C entry (QCE1) --> C entry (QCE2) --> C function (qce2a)     */
/*      intra module --> return all the way back                     */
/*   2: C entry (QCE2) --> C function (qsc2a) inter module and pass  */
/*      an intra module function pointer --> call passed function    */
/*      (qsc2b) --> return all the way back                          */
/*   3: C entry (QCE3) --> C function (qsc1b) inter module and pass  */
/*      an inter module function pointer from a 3rd module --> call  */
/*      passed function (qsc3a) --> return all the way back          */
/*   4: C entry (QCE1) --> C function (qce1a) intra module and pass  */
/*      an inter module function pointer --> call passed function    */
/*      (qsc1a) --> return all the way back                          */
/*   5: C (main) -> C function (intra) and pass an intra module      */
/*      function pointer -> call passed function -> return all the   */
/*      way back                                                     */
/*   6: C (main) -> C entry -> C function (intra) -> C function      */
/*      (inter) -> return all the way back                           */
/*   7: C (main) -> C function (intra) -> C entry -> C function      */
/*      (inter) --> return all the way back                          */
/*   8: C (main) -> C function (inter) -> C entry -> return all the  */
/*      way back                                                     */
/*   9: C (main) -> C function (inter) and pass an intra module      */
/*      function pointer -> call passed function -> return all the   */
/*      way back                                                     */
/*  10: C (main) -> C function (inter) and pass an inter module      */
/*      function pointer -> call passed function -> return all the   */
/*      way back                                                     */
/*  11: Testcase to test User Defined Errno's   C-021 3193f          */
/*  12: Test ENTRC with new options          C-031 3503f             */
/*  13: Test long function names                                     */
/*  14: Test very long function names (greater than 2000 characters) */         //CRK7298
/*  15: Test BSO calls QQQQ                                          */         //f4692
/*  16: Test CSO calls QQQQ                                          */         //f4692
/*  17: Test BSO calls data program QXAX                             */         //f4692
/*  18: Test CSO calls data program QXAX                             */         //f4692
/*  19: Test BSO calls GTAL                                          */         //f4692
/*  20: Test CSO calls GTAL                                          */         //f4692
/*  21: Test CSO calls non-existent function                         */         //f4692
/*  22: Run test case 15-21                                          */         //f4692
/*  23-25: Unused                                                    */         //f4692
/*                                                                   */
/* BAL test cases: (test case slots 26-50)                           */
/*  26: BAL (QSBA) --> BAL (QSBB) intra module --> BAL (QSBF) inter  */
/*        module, base-less --> BAL (QSBC) inter module TV --> return*/
/*        all the way back.                                          */
/*  27: BAL31 (QSBA) --> BAL31 (QSBB) intra module, >4k, >1 base reg */
/*        --> BAL31 (QSBE) inter module --> BAL31 (QSBD) inter       */
/*        module TV --> return all the way back                      */
/*  28: BAL64 (QSB1) --> BAL64 (QSB2) intra module --> BAL64 (QSB6)  */
/*        inter module TV --> BAL64 (QSB9) inter module, baseless -->*/
/*        return all the way back                                    */
/*  29: BAL64 (QSB1) --> BAL64 (QSB4) intra module, >4k, > 1 base reg*/
/*        --> BAL64 (QSB6) inter module TV --> BAL64 (QSB8) inter    */
/*        module --> return all the way back                         */
/*  30: BAL (QSBC) --> BAL64 (QSB1) entry --> BAL64 (QSB2) intra     */
/*        module TV --> BAL64 (QSB6) inter module TV --> return all  */
/*        the way back.                                              */
/*  31: BAL64 (QSB5) --> BAL (QSBC) TV --> BAL64 (QSB7) inter module */
/*        --> BAL64 (QSBF) inter module --> BAL64 (QSB8) intra module*/
/*        --> return all the way back                                */
/*  32: BAL (QSBE) --> ENTDC to BAL (QSBD) inter module --> ENTNC to */
/*        BAL (QSBC) intra module TV --> exit                        */
/*  33: BAL64 (QSB1) ENTDC to BAL64 (QSB3) intra module TV ENTNC to  */
/*        BAL64 (QSB6) inter module TV --> exit                      */
/*  34:                                                              */
/*  35-50: Unused                                                    */
/*                                                                   */
/*                                                                   */
/* Mixed C and BAL test cases: (test case slots 51-75)               */
/*  51: C entry (QCE1) --> BAL64 entry (QSB8) --> C entry (QCE2) --> */
/*       BAL entry (QSBC) --> return all the way back                */
/*  52: C entry (QCE3) --> BAL (QSBF) --> C entry (QCE1) --> BAL64   */
/*       (QSB6) --> return all the way back                          */
/*  53: BAL entry (QSBA) --> C entry (QCE2) --> BAL64 (QSB6) TV -->  */
/*       C entry (QCE3) --> BAL (QSBD) TV --> return all the way back*/
/*  54: BAL64 (QSB1) --> C entry (QCE2) --> BAL64 (QSB6) TV -->      */
/*       C entry (QCE3) --> BAL (QSBD) TV --> return all the way back*/
/*  55: C main (QME1) --> BAL64 (QSB7) --> C entry (QCE3) --> BAL    */
/*       (QSBF) --> return all the way back                          */
/*  56: C main (QME2) --> BAL (QSBF) --> C entry (QCE1) --> BAL64    */
/*       (QSB8) --> return all the way back                          */
/*  57: C main (QME3) --> BAL64 (QSB6) TV --> C entry (QCE3) -->     */
/*       BAL (QSBF) --> return all the way                           */
/*  58: BAL64 (QSB3) --> CALLC to C function (qsc3a) --> BAL (QSBB)  */
/*       --> return all the way back                                 */
/*  59: BAL64 (QSB4) --> CALLC to C function (qsc2b) --> BAL64 (QSB9)*/
/*       --> return all the way back                                 */
/*  60: BAL (QSBE) --> CALLC to C function (qsc3b) --> C entry (QCE3)*/
/*       --> return all the way back                                 */
/*  61:                                                              */
/*  62:C entry (QCE3) --> BAL64 (QSB1) --> C entry (QCE1) --> BAL64  */
/*      (QSB5)                                                       */
/*  63:C entry (QCE3) --> BAL64 (QSB1) --> C entry (QCE1) --> BAL64  */
/*      (QSB5) --> C entry (QCE2)                                    */
/*  64-75: Unused                                                    */
/*                                                                   */
/*                                                                   */
/* Function Trace test cases: (test case slots 100 - 123)            */
/*  100: C++ main (QME4) -> various C++ functions                    */
/*  101: C main (QME1) -> functions to test char                     */
/*  102: C main (QME1) -> functions to test unsigned char            */
/*  103: C main (QME1) -> functions to test short                    */
/*  104: C main (QME1) -> functions to test unsigned short           */
/*  105: C main (QME1) -> functions to test int                      */
/*  106: C main (QME1) -> functions to test unsigned int             */
/*  107: C main (QME1) -> functions to test long                     */
/*  108: C main (QME1) -> functions to test unsigned long            */
/*  109: C main (QME1) -> functions to test long long                */
/*  110: C main (QME1) -> functions to test unsigned long long       */
/*  111: C main (QME1) -> functions to test float                    */
/*  112: C main (QME1) -> functions to test double                   */
/*  113: C main (QME1) -> functions to test long double              */
/*  114: C main (QME1) -> functions to test variable parameters      */
/*  115: C main (QME1) -> functions to test void parameters          */
/*  116: C main (QME1) -> functions to test tpf_trace_info           */
/*  117: C main (QME1) -> functions to test tpf_trace_info           */
/*  118: C main (QME1) -> functions to test tpf_trace_info           */
/*  119: C main (QME1) -> functions to test tpf_trace_info           */
/*  120: C main (QME1) -> functions to test tpf_trace_info           */
/*  121: C main (QME1) -> functions to test typedef                  */
/*  122: C main (QME1) -> functions to test structure in register    */
/*  123: C main (QME1) -> functions to test structure in register    */
/*                                                                   */
/*                                                                   */
/* NOTE:  See block comments in code for linkage details             */
/*                                                                   */
/* CHANGE ACTIVITY -  (also see PLOG activity below)                 */
/*                                                                   */
/*                                                                   */
/*********************************************************************/
/*                                                                   */
/* DEPENDENCIES... None.                                             */
/*                                                                   */
/* RESTRICTIONS... None.                                             */
/*                                                                   */
/* SPECIAL CONSIDERATIONS...                                         */
/*                                                                   */
/* INTERFACES:                                                       */
/*                                                                   */
/* For all of the segments in SOLD, the test case number will        */
/* be saved in the ECB in EBW000.  This is done in order to handle,  */
/* in particular, the need to send the test case number while doing  */
/* an ENTDC, which clears the stack when entering a C segment.       */
/*                                                                   */
/* For continuous mode, QSCB sets up an ECB workarea to pass         */
/* pertinent information to QSCC via CREMC.                          */
/* The workarea is off data level 0 (D0).  This workarea starts at   */
/* EBW000.                                                           */
/*                                                                   */
/* -  test case number      (EBW000)                                 */
/* -  terminal address      (EBROUT)                                 */
/* -  CPUID                 (CEICPD)                                 */
/* -  mode continuous flag  (EBW008)                                 */
/*                                                                   */
/*********************************************************************/

#include <stdlib.h>           /* for exit                             */
#include <stdio.h>
#include <string.h>           /* for memcpy,strcmp,strpbrk,strchr     */
#include <errno.h>
#include <sys/types.h>

#include <tpf/c_eb0eb.h>      /* to address the ECB                   */
#include <tpf/tpfapi.h>       /* for cinfc, wtopc                     */
#include <tpf/tpfparse.h>     /* for IPRSE                            */
#include <tpf/sysapi.h>

#include <c_sold.h>

static void start_continuous_mode();
static void stop_continuous_mode();

/**********************************************************************/
/* end of #include and #define                                        */
/**********************************************************************/


void QSCA(void)
{
char   *state_ind;                  /* pointer to the state indicator       */
char    header_buffer[BUFNO];       /* fill in text for output msg          */
struct  mi0mi *block_ptr;           /* pointer to message block             */
struct  IPRSE_output parsed_struc;  /* pointer to parse msg                 */
struct  IPRSE_output *parse_ptr;    /* pointer to parse output              */
char   *input_ptr;                  /* pointer to input message text        */
short   result;                     /* return code from IPRSE               */

struct pat *pat_addr;                                                      //d10433f_1
char prog_name[PAT_PROG_NAME_LEN+PAT_VERS_CODE_LEN+1] = {0};               //d10875

/**********************************************************************/
/* Using IPRSE to check the syntax of the input function message:     */
/*   ZTEST SOLD  < START|CASE-xx|STATUS|STOP >                        */
/*                           xx is the test case number               */
/**********************************************************************/
char *grammar_ptr = "{ START "                                         \
                    "| CASE-d++ [(NO)DUMP]"                            \
                    "| CASE-d++ [(NO)SHOWVER]"                         \
                    "| STATUS "                                        \
                    "| STOP "                                          \
                    "}";


/*--------------------------------------------------------------------*/
/* Help info follows                                                  */
/*--------------------------------------------------------------------*/
char *help_info1 = "SOLD SYNTAX:\n"                                    \
          "  ZTEST i SOLD CASE-xx ((NO)DUMP|(NO)SHOWVER)\n"            \
          "     Execute one test case.\n"                              \
          "  ZTEST i SOLD START\n"                                     \
          "     Start continuous operation.\n"                         \
          "  ZTEST i SOLD STATUS\n"                                    \
          "     Display status of running driver.";

char *help_info2 =
          "  ZTEST i SOLD STOP\n"                                      \
          "     Stop SOLD in continuous mode\n"                        \
          "\nPARAMETERS:\n"                                            \
          "  i  : I-STREAM NUMBER\n"                                   \
          "  xx : TEST CASE NUMBER, 1-125\n"                           \
          "  DUMP : CAUSE THE DRIVER TO DUMP AFTER COMPLETION";

char *help_info3 =
       "  SHOWVER : DISPLAY THE PROGRAM VERSION AND VARIATION NUMBER\n"\
          "\nEND OF DISPLAY";

/*--------------------------------------------------------------------*/
/* Invalid syntax used message follows                                */
/*--------------------------------------------------------------------*/
char *sold0099e = "You entered ZTEST SOLD with invalid syntax.\n"
                  "  Enter ZTEST SOLD HELP to display the correct syntax";

/*--------------------------------------------------------------------*/
/* Setup  the input message pointers. Including the pointer to the    */
/* input message. This pointer has been bumped to the first parameter.*/
/*--------------------------------------------------------------------*/

block_ptr = (struct mi0mi *) ecbptr()->ce1cr0;
input_ptr = block_ptr->mi0acc;
input_ptr = strpbrk(input_ptr,"SOLD");
input_ptr += 4;

/*--------------------------------------------------------------------*/
/* Initialize the output structure of the input array.                */
/* Enter the parser. With the return code in result.                  */
/*--------------------------------------------------------------------*/

result = IPRSE_parse(input_ptr, grammar_ptr, &parsed_struc,            \
                     IPRSE_ALLOC+IPRSE_NOPRINT+IPRSE_EOM,              \
                     NULL);
parse_ptr = &parsed_struc;

/*--------------------------------------------------------------------*/
/* If result = 0 (keywords not found) then respond with the invalid   */
/* syntax error message and processing stop.                          */
/*--------------------------------------------------------------------*/
 if (result == 0)
 {
  wtopc_insert_header(header_buffer, "SOLD", 99, 'E',
                      WTOPC_SYS_TIME);
  wtopc(sold0099e, WTOPC_EBROUT, WTOPC_NO_CHAIN, header_buffer);
  exit(0);
 }

/*--------------------------------------------------------------------*/
/* If result < 0 (? or help entered as a keyword) then respond with   */
/* help info and exit.                                                */
/*--------------------------------------------------------------------*/

 if (result < 0)
 {
  wtopc_insert_header(header_buffer, "SOLD", 10, 'I', WTOPC_SYS_TIME);
  wtopc(help_info1, 0, WTOPC_CHAIN, header_buffer);
  wtopc(help_info2, 0, WTOPC_CHAIN, WTOPC_NO_HEADER);
  wtopc(help_info3, 0, WTOPC_CHAIN_END, WTOPC_NO_HEADER);
  exit(0);
 }

/**********************************************************************/
/* SOLD can be run in 1052 or NORM state.  The following              */
/* code is to check the system state indicator CMMSTI.                */
/* SOLD stops if system is cycling.                                   */
/**********************************************************************/
 state_ind=cinfc(CINFC_READ,CINFC_CMMSTI);
 if (state_ind[1] != 0)
         {
          wtopc_insert_header(header_buffer, "SOLD", 98, 'E', WTOPC_SYS_TIME);
          wtopc("SYSTEM IS CYCLING - SOLD CANNOT BE RUN", 0,
            WTOPC_NO_CHAIN, header_buffer);
          exit(0);
         }


 /* Check which parameter was entered */
 if (!strcmp(parsed_struc.IPRSE_parameter,"START")) {

   /****************************************************/
   /* Indicate start of SOLD driver in continuous mode */
   /****************************************************/
   start_continuous_mode();                                                     //C64CF

 } else if (!strcmp(parsed_struc.IPRSE_parameter,"CASE-d++")) {

   /**********************************************************************/
   /* If the parameter is CASE-xx, xx is the test case number to be      */
   /* executed.                                                          */
   /**********************************************************************/
   test_case = atoi(parse_ptr->IPRSE_value);
   test_mode = CONT_NO;

   /* If DUMP was specified, set test_mode to DUMP_YES.  This will cause   */
   /* the driver to dump and prevent start and stop messages from printing */
   if (result == 2) {
      parse_ptr = parse_ptr->IPRSE_next;                                        //C64JK

      if ( strcmp(parse_ptr->IPRSE_parameter, "(NO)DUMP") == 0 )                   //f5555
      {
         if( !strcmp( parse_ptr->IPRSE_value, "Y") )                               //C64JK
         {
            /* Dump option is not valid for case 32 and case 33. */
            if (test_case == 32 || test_case == 33)
            {
               wtopc_insert_header(header_buffer, "SOLD", 150, 'E', WTOPC_SYS_TIME);
               wtopc("DUMP OPTION IS INVALID FOR TEST CASE 32 OR 33.", 0,
                 WTOPC_NO_CHAIN, header_buffer);
               exit(0);
            }
            else
               test_mode = DUMP_YES;                                               //C64JK
         }
      }

      else if ( strcmp(parse_ptr->IPRSE_parameter, "(NO)SHOWVER") == 0 )           //f5555
      {                                                                            //f5555
         if( !strcmp( parse_ptr->IPRSE_value, "Y") )                               //f5555
         {                                                                         //f5555
            if (test_case < 23 && test_case > 14) //Loader Dynamic Linkage cases   //f5555
            {                                                                      //f5555
               wtopc_insert_header(header_buffer, "SOLD", 160, 'E', WTOPC_SYS_TIME);//f5555
               wtopc("SHOWVER OPTION IS INVALID FOR TEST CASE 15-22.", 0,          //f5555
                 WTOPC_NO_CHAIN, header_buffer);                                   //f5555
               exit(0);                                                            //f5555
            }                                                                      //f5555
            else                                                                   //f5555
               test_mode = SHOW_VER;                                               //f5555
               pat_addr = (void*) tpf_ecbp3()->ce3patg;                            //f12472
               memcpy(prog_name, pat_addr->patname, PAT_PROG_NAME_LEN+PAT_VERS_CODE_LEN);//d10875
               show_prog_version(prog_name, pat_addr->patacnum);                         //f12305
         }
      }
   }  //if (result == 2) {                                                         //f5555

   SOLD_do_test_case();                                                            //C64JK

 } else if (!strcmp(parsed_struc.IPRSE_parameter,"STOP")) {

   /**********************************************************************/
   /*                                                                    */
   /* If the parameter is STOP, driver SOLD will stop.                   */
   /*                                                                    */
   /**********************************************************************/
   stop_continuous_mode();                                                      //C64CF

 } else if (!strcmp(parsed_struc.IPRSE_parameter,"STATUS")) {

   /**********************************************************************/
   /*                                                                    */
   /* If the parameter is STATUS, display status of the driver.          */
   /*                                                                    */
   /**********************************************************************/
   status_msg();                                                                //C64JK
 }                           /* End 'if'     */

 exit(0);                                                                       //C64CF
}


static void start_continuous_mode()                                             //C64CF
{                                                                               //C64CF
   long   size;                                                                 //C64CF
   char   header_buffer[BUFNO];                                                 //C64CF
   char   message_buf[BUFNO];     /* fill in text for output msg           */   //F9127
   struct sold_cntl *ctl_ptr = NULL;                                            //C64CF

   char   sold_token[9];                                                        //F11565
   THIS_SS_SOLD_CNTL_TOKEN(sold_token);                                         //F11565

   while (ctl_ptr == NULL) {                                                    //C64CF
      /* See if a control block already exists */
      ctl_ptr = (struct sold_cntl *)tpf_fsysc(sold_token, &size);               //C64CF

      if (ctl_ptr == NULL) {                                                    //C64CF
         /* No control block - request one */
         ctl_ptr = (struct sold_cntl *)tpf_gsysc(1, sold_token, NULL,           //C64CF
                                           GSYSC_UNIQUE + GSYSC_64BIT);         //C64CF

         if (ctl_ptr != NULL) {                                                 //C64CF
           /* A block has been allocated.  Intialize it */
           /* Turn on storage protection override */
           tpf_stpoc(TPF_STPOC_ON, NULL);                                       //C64CF

           /* Clear the control block */
           memset(ctl_ptr, 0, sizeof(struct sold_cntl));                        //C64CF

           /* Lock the control block */
           lockc(&(ctl_ptr->lock), LOCK_O_SPIN);                                //C64CF

           /* Initialize the header and status */
           ctl_ptr->IS_stat[(ecbptr()->ce1isn)-1].status = START;               //F9127
           memcpy(&(ctl_ptr->hdr), SOLD_HDR, sizeof(ctl_ptr->hdr));             //C64CF

           /* Unlock the control block and turn off the override */
           unlkc(&(ctl_ptr->lock));                                             //C64CF
           tpf_stpoc(TPF_STPOC_OFF, NULL);                                      //C64CF

         } else if ((ctl_ptr == NULL) && (errno == ETPFGSYS_NUNQTKN)) {         //C64CF
           /* A NULL was returned and the error is due to       */
           /* a block by that token already existing.  Just go  */
           /* back to the top and try another fsysc.            */
           continue;                                                            //C64CF

         } else {                                                               //C64CF
           /* A NULL pointer was returned and the errno was not */
           /* due to another i-stream allocating the same block */
           /* first.  This is a system-gsysc error.  Issue an   */
           /* error message and exit.                           */
           wtopc_insert_header(header_buffer, "SOLD", 88, 'E', WTOPC_SYS_TIME); //C64CF
           wtopc("ERROR GETTING SYSTEM HEAP FOR CONTROL BLOCK", WTOPC_EBROUT,   //C64CF
               WTOPC_NO_CHAIN, header_buffer);                                  //C64CF
           exit(-1);                                                            //C64CF
         }                                                                      //C64CF

      } else {                                                                  //C64CF
         /* The fsysc returned a block.  First check the header */
         if (memcmp(&(ctl_ptr->hdr), SOLD_HDR, sizeof(ctl_ptr->hdr)) == 0) {    //C64CF

            /* The header is inialized.  Check the status */
            if (ctl_ptr->IS_stat[(ecbptr()->ce1isn)-1].status == STOP) {        //F9127
               /* The status is STOP - do not start another instance */
               sprintf(message_buf, "UNABLE TO START - SOLD IS STOPPING ON ISTREAM %d",
                                                 ecbptr()->ce1isn);             //F9127
               wtopc_insert_header(header_buffer, "SOLD", 87, 'E', WTOPC_SYS_TIME);
               wtopc(message_buf, WTOPC_EBROUT, WTOPC_NO_CHAIN, header_buffer); //C64CF
               exit(-1);                                                        //C64CF
            }                                                                   //C64CF
         } else {                                                               //C64CF
            /* The header is not initialized.  Defer and redo the fsysc */
            /* We do the fsysc again to take care of possible race      */
            /* conditions with ZTEST SOLD STOP.                         */
            ctl_ptr = NULL;                                                     //C64CF
            defrc();                                                            //C64CF
         }                                                                      //C64CF
      }                                                                         //C64CF
   } /* end while */                                                            //C64CF


   sprintf(message_buf, "SOLD IS STARTING IN CONTINUOUS MODE ON ISTREAM %d",
                           ecbptr()->ce1isn);                                   //F9127
   wtopc_insert_header(header_buffer, "SOLD", 1, 'I', WTOPC_SYS_TIME);          //C64JK
   wtopc(message_buf, WTOPC_EBROUT, WTOPC_NO_CHAIN, header_buffer);             //C64CF

   /* Increment the driver counts for a new instance of SOLD */
   tpf_stpoc(TPF_STPOC_ON, NULL);                                               //C64CF
   lockc(&(ctl_ptr->lock), LOCK_O_SPIN);                                        //C64CF
   ctl_ptr->IS_stat[(ecbptr()->ce1isn)-1].status = START;                       //F9127
   ctl_ptr->total_cnt++;                                                        //C64CF
   ctl_ptr->IS_stat[(ecbptr()->ce1isn)-1].drvr_cnt++;                           //C64CF
   unlkc(&(ctl_ptr->lock));                                                     //C64CF
   tpf_stpoc(TPF_STPOC_OFF, NULL);                                              //C64CF

   QSCB();                   /* Invoke QSCB to do continuous mode     */        //C64JK

   return;                                                                      //C64CF
}                                                                               //C64CF


static void stop_continuous_mode()                                              //C64CF
{                                                                               //C64CF
   long   size;                                                                 //C64CF
   char   header_buffer[BUFNO];                                                 //C64CF
   char   message_buf[BUFNO];     /* fill in text for output msg           */   //F9127
   struct sold_cntl *ctl_ptr;                                                   //C64CF

   char   sold_token[9];                                                        //F11565
   THIS_SS_SOLD_CNTL_TOKEN(sold_token);                                         //F11565

   ctl_ptr = (struct sold_cntl *)tpf_fsysc(sold_token, &size);                  //C64CF
   if (ctl_ptr == NULL) {                                                       //C64CF
      /* Control pointer not found.  SOLD was never started */
      wtopc_insert_header(header_buffer, "SOLD", 95, 'E', WTOPC_SYS_TIME);      //C64CF
      wtopc("SOLD IS NOT RUNNING", WTOPC_EBROUT,                                //C64CF
              WTOPC_NO_CHAIN, header_buffer);                                   //C64CF
      exit(0);                                                                  //C64CF
   }                                                                            //C64CF

   if (ctl_ptr->IS_stat[(ecbptr()->ce1isn)-1].status == INIT) {                 //C64CF
      /* Control pointer not found.  SOLD was never started */
      sprintf(message_buf, "SOLD IS NOT RUNNING ON ISTREAM %d", ecbptr()->ce1isn);
      wtopc_insert_header(header_buffer, "SOLD", 95, 'E', WTOPC_SYS_TIME);      //C64CF
      wtopc(message_buf, WTOPC_EBROUT, WTOPC_NO_CHAIN, header_buffer);          //C64CF
      exit(0);                                                                  //C64CF
   }                                                                            //C64CF

   /* To avoid race conditions with START, make sure the header is */
   /* initialized before trying the STOP operation                 */
   while (memcmp(&(ctl_ptr->hdr), SOLD_HDR, sizeof(ctl_ptr->hdr)) != 0) {       //C64CF
      defrc();                                                                  //C64CF
   }                                                                            //C64CF

   /* Stop the SOLD driver */
   tpf_stpoc(TPF_STPOC_ON, NULL);                                               //C64CF
   lockc(&(ctl_ptr->lock), LOCK_O_SPIN);                                        //C64CF
   if (ctl_ptr->IS_stat[(ecbptr()->ce1isn)-1].status == STOP)                   //F9127
   {
      /* SOLD already stopping.  Issue error */
      unlkc(&(ctl_ptr->lock));                                                  //C64CF
      tpf_stpoc(TPF_STPOC_OFF, NULL);                                           //C64CF
      sprintf(message_buf, "SOLD IS ALREADY STOPPING ON ISTREAM %d", ecbptr()->ce1isn);
      wtopc_insert_header(header_buffer, "SOLD", 97, 'E', WTOPC_SYS_TIME);      //C64CF
      wtopc(message_buf, WTOPC_EBROUT, WTOPC_NO_CHAIN, header_buffer);          //C64CF
      exit(0);                                                                  //C64CF
   } else {                                                                     //C64CF
      ctl_ptr->IS_stat[(ecbptr()->ce1isn)-1].status = STOP;                     //F9127
      unlkc(&(ctl_ptr->lock));                                                  //C64CF
      tpf_stpoc(TPF_STPOC_OFF, NULL);                                           //C64CF
   }                                                                            //C64CF

   /* Issue message - starting the STOP process */
   sprintf(message_buf, "STOPPING ALL INSTANCES OF SOLD ON ISTREAM %d",
                                         ecbptr()->ce1isn);                     //F9127
   wtopc_insert_header(header_buffer, "SOLD", 5, 'I', WTOPC_SYS_TIME);          //C64CF
   wtopc(message_buf, WTOPC_EBROUT, WTOPC_NO_CHAIN, header_buffer);             //C64CF

   /* Defer until all SOLD drivers on this istream have exited */
   while (ctl_ptr->IS_stat[(ecbptr()->ce1isn)-1].drvr_cnt > 0) {                //F9127
      defrc();                                                                  //C64CF
   }                                                                            //C64CF

   /* Reset the status bit so user can start it again if necessary. */
   tpf_stpoc(TPF_STPOC_ON, NULL);                                               //F9127
   ctl_ptr->IS_stat[(ecbptr()->ce1isn)-1].status = INIT;                        //F9127
   tpf_stpoc(TPF_STPOC_OFF, NULL);                                              //F9127

   /* If all instances of SOLD drivers are gone, release the system */
   /* heap frame used by SOLD driver.                               */
   if (ctl_ptr->total_cnt == 0)                                                 //F9127
   {
      /* Clear the header in case same block is reused on next start */
      tpf_stpoc(TPF_STPOC_ON, NULL);                                            //C64CF
      memset(&(ctl_ptr->hdr), 0, sizeof(ctl_ptr->hdr));                         //C64CF
      tpf_stpoc(TPF_STPOC_OFF, NULL);                                           //C64CF

      /* Release the control block */
      tpf_rsysc(ctl_ptr, 1, sold_token);                                        //C64CF
   }

   /* Issue message - stop complete */
   sprintf(message_buf, "SOLD STOPPED ON ISTREAM %d", ecbptr()->ce1isn);        //F9127
   wtopc_insert_header(header_buffer, "SOLD", 6, 'I', WTOPC_SYS_TIME);          //C64CF
   wtopc(message_buf, WTOPC_EBROUT, WTOPC_NO_CHAIN, header_buffer);             //C64CF

   return;                                                                      //C64CF
}                                                                               //C64CF
